// MyExt.cpp

#include "dbgexts.h"

BOOL PreferDML(PDEBUG_CLIENT pDebugClient)
{
	BOOL bPreferDML = FALSE;
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)& pDebugControl)))
	{
		ULONG ulOptions = 0;
		if (SUCCEEDED(pDebugControl->GetEngineOptions(&ulOptions)))
		{
			bPreferDML = (ulOptions & DEBUG_ENGOPT_PREFER_DML);
		}
		pDebugControl->Release();
	}
	return bPreferDML;
}

BOOL AbilityDML(PDEBUG_CLIENT pDebugClient)
{
	BOOL bAbilityDML = FALSE;
	IDebugAdvanced2* pDebugAdvanced2;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugAdvanced2), (void **)& pDebugAdvanced2)))
	{
		HRESULT hr = 0;
		if (SUCCEEDED(hr = pDebugAdvanced2->Request(DEBUG_REQUEST_CURRENT_OUTPUT_CALLBACKS_ARE_DML_AWARE, NULL, 0, NULL, 0, NULL)))
		{
			if (hr == S_OK) bAbilityDML = TRUE;
		}
		pDebugAdvanced2->Release();
	}
	return bAbilityDML;
}

HRESULT CALLBACK 
hellodml(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_NORMAL, "<b>Hello</b> <i>DML</i> <u>World!</u>\n");
		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
echoasdml(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "[Start DML]\n");
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_NORMAL, "%s\n", args);
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "[End DML]\n");
		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
echoasdmlescape(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "[Start DML]\n");
		if ((args != NULL) && (strlen(args) > 0))
		{
			char* szEscape = (char*)malloc(strlen(args) * 6);
			if (szEscape == NULL)
			{
				pDebugControl->Release();
				return E_OUTOFMEMORY;
			}
			size_t n=0; size_t e=0;
			for (; n<strlen(args); n++)
			{
				switch (args[n])
				{
				case '&':
					memcpy(&szEscape[e], "&amp;", 5);
					e+=5;
					break;
				case '<':
					memcpy(&szEscape[e], "&lt;", 4);
					e+=4;
					break;
				case '>':
					memcpy(&szEscape[e], "&gt;", 4);
					e+=4;
					break;
				case '"':
					memcpy(&szEscape[e], "&quot;", 6);
					e+=6;
					break;
				default:
					szEscape[e] = args[n];
					e+=1;
					break;
				}
			}
			szEscape[e++] = '\0';
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_NORMAL, "%s\n", szEscape);
			free(szEscape);
		}
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "[End DML]\n");
		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
memorydml(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	BOOL bDone = FALSE;

	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		IDebugSymbols* pDebugSymbols;
		if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugSymbols), (void **)&pDebugSymbols)))
		{
			IDebugDataSpaces* pDebugDataSpaces;
			if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pDebugDataSpaces)))
			{
				// Resolve the symbol
				ULONG64 ulAddress = 0;
				if ((args != NULL) && (strlen(args) > 0) && SUCCEEDED(pDebugSymbols->GetOffsetByName(args, &ulAddress)))
				{	// Read the value of the pointer from the target address space
					ULONG64 ulPtr = 0;
					if (SUCCEEDED(pDebugDataSpaces->ReadPointersVirtual(1, ulAddress, &ulPtr)))
					{
						char szBuffer[256];
						ULONG ulBytesRead = 0;
						if (SUCCEEDED(pDebugDataSpaces->ReadVirtual(ulPtr, szBuffer, 255, &ulBytesRead)))
						{
							szBuffer[ulBytesRead] = '\0';

							// Output the value via %ma and %s
							pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_TEXT, DEBUG_OUTPUT_NORMAL, "[Start DML]\n");
							pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_ERROR, "<b>Error</b> (  %%ma): %ma\n", ulPtr);
							pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_ERROR, "<b>Error</b> (%%Y{t}): %Y{t}\n", szBuffer);
							pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_ERROR, "<b>Error</b> (   %%s): %s\n", szBuffer);
							pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_TEXT, DEBUG_OUTPUT_NORMAL, "[End DML]\n");
							bDone = TRUE;
						}
					}
				}
				pDebugDataSpaces->Release();
			}
			pDebugSymbols->Release();
		}
		if (!bDone)
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_TEXT, DEBUG_OUTPUT_EXTENSION_WARNING, "Unable to output memory for symbol '%s'\n", args);

		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
ifdml(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	PDEBUG_CONTROL pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		// A condition is usually not required;
		// Rely on content conversion when there isn't any abbreviation or superfluous content
		if (PreferDML(pDebugClient) && AbilityDML(pDebugClient))
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_NORMAL, "<b>Hello</b> <i>DML</i> <u>World!</u>\n");
		}
		else
		{
			pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_TEXT, DEBUG_OUTPUT_NORMAL, "Hello TEXT World!\n");
		}
		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT CALLBACK 
maskdml(PDEBUG_CLIENT pDebugClient, PCSTR args)
{
	UNREFERENCED_PARAMETER(args);

	PDEBUG_CONTROL pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_NORMAL, "<b>DEBUG_OUTPUT_NORMAL</b> - Normal output.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_ERROR, "<b>DEBUG_OUTPUT_ERROR</b> - Error output.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_WARNING, "<b>DEBUG_OUTPUT_WARNING</b> - Warnings.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_VERBOSE, "<b>DEBUG_OUTPUT_VERBOSE</b> - Additional output.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_PROMPT, "<b>DEBUG_OUTPUT_PROMPT</b> - Prompt output.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_PROMPT_REGISTERS, "<b>DEBUG_OUTPUT_PROMPT_REGISTERS</b> - Register dump before prompt.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_EXTENSION_WARNING, "<b>DEBUG_OUTPUT_EXTENSION_WARNING</b> - Warnings specific to extension operation.\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_DEBUGGEE, "<b>DEBUG_OUTPUT_DEBUGGEE</b> - Debug output from the target (for example, OutputDebugString or DbgPrint).\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_DEBUGGEE_PROMPT, "<b>DEBUG_OUTPUT_DEBUGGEE_PROMPT</b> - Debug input expected by the target (for example, DbgPrompt).\n");
 		pDebugControl->ControlledOutput(DEBUG_OUTCTL_AMBIENT_DML, DEBUG_OUTPUT_SYMBOLS, "<b>DEBUG_OUTPUT_SYMBOLS</b> - Symbol messages (for example, !sym noisy).\n");
		pDebugControl->Release();
	}
	return S_OK;
}
